import os
import json
import itertools
import numpy as np
from typing import Dict, List, Any, Optional, Tuple


class Config:
    """
    Abstract base class for model configurations.
    It defines hyperparameter search space
    for a model.

    Supported methods:
    - generate_hyperparams_random
        Generate hyperparameter combinations
        for random search.
    """
    PARAMS_DICT: Dict[str, List] = {
        'sample_param': ['sample_value1', 'sample_value2']
    }

    @classmethod
    def generate_hyperparams_random(
            cls,
            n_trials: int = 1,
            dir: Optional[str] = None) -> Tuple[Dict, str]:
        """
        Generate hyperparameter combinations for random trials.

        Parameters
        ----------
        n_trials: int
            Number of trials for random search.
        dir: Optional[str]
            Directory path to save json file of generated
            combinations.

        Returns
        -------
        trials_dict: Dict
            Dictionary of combinations generated for trials.
        file_path: str
            File path of saved json file containing generated
            combinations.
        """
        hyperparameter_combs: List[Dict[
            str, Any]] = cls._generate_random_hyperparam_values(n=n_trials)

        trials_dict: Dict = {}
        for count, params in enumerate(hyperparameter_combs):
            trials_dict[f'trial_{count+1}'] = params

        file_name: str = f"{n_trials}_trials_params.json"
        if dir is None:
            cwd: str = os.getcwd()
            file_path: str = os.path.join(cwd, file_name)
        else:
            file_path = os.path.join(dir, file_name)

        with open(file_path, "w") as json_file:
            json.dump(trials_dict, json_file, indent=4)

        return trials_dict, file_path

    @classmethod
    def _generate_random_hyperparam_values(cls,
                                           n: int) -> List[Dict[str, Any]]:
        """
        Generates `n` random hyperparameter combinations
        of hyperparameter values

        Parameters
        ----------
        n: int
            Number of random combinations to generate

        Returns
        -------
        params_subset: List[Dict[str, Any]]
            list of hyperparameter combinations
        """
        hyperparam_keys: List
        hyperparam_values: List
        hyperparam_keys, hyperparam_values = [], []
        for key, values in cls.PARAMS_DICT.items():
            if callable(values):
                # If callable, sample it for a maximum n times
                values = [values() for i in range(n)]
            hyperparam_keys.append(key)
            hyperparam_values.append(values)

        hyperparam_combs: List = []
        for iterable_hyperparam_comb in itertools.product(*hyperparam_values):
            hyperparam_comb: List = list(iterable_hyperparam_comb)
            hyperparam_combs.append(hyperparam_comb)
        indices: np.ndarray = np.random.permutation(len(hyperparam_combs))[:n]

        params_subset: List = []
        for index in indices:
            param: Dict = {}
            for key, hyperparam_value in zip(hyperparam_keys,
                                             hyperparam_combs[index]):
                param[key] = hyperparam_value
            params_subset.append(param)
        return params_subset



import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import List, Tuple, Union, Optional, Callable, Dict

from deepchem.models.losses import Loss, L2Loss
from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models.optimizers import Optimizer, LearningRateSchedule

from openpom.layers.pom_ffn import CustomPositionwiseFeedForward
from openpom.utils.loss import CustomMultiLabelLoss
from openpom.utils.optimizer import get_optimizer

try:
    import dgl
    from dgl import DGLGraph
    from dgl.nn.pytorch import Set2Set
    from openpom.layers.pom_mpnn_gnn import CustomMPNNGNN
except (ImportError, ModuleNotFoundError):
    raise ImportError('This module requires dgl and dgllife')


class MPNNPOM(nn.Module):
    """
    MPNN model computes a principal odor map
    using multilabel-classification based on the pre-print:
    "A Principal Odor Map Unifies DiverseTasks in Human
        Olfactory Perception" [1]

    This model proceeds as follows:

    * Combine latest node representations and edge features in
        updating node representations, which involves multiple
        rounds of message passing.
    * For each graph, compute its representation by radius 0 combination
        to fold atom and bond embeddings together, followed by
        'set2set' or 'global_sum_pooling' readout.
    * Perform the final prediction using a feed-forward layer.

    References
    ----------
    .. [1] Brian K. Lee, Emily J. Mayhew, Benjamin Sanchez-Lengeling,
        Jennifer N. Wei, Wesley W. Qian, Kelsie Little, Matthew Andres,
        Britney B. Nguyen, Theresa Moloy, Jane K. Parker, Richard C. Gerkin,
        Joel D. Mainland, Alexander B. Wiltschko
        `A Principal Odor Map Unifies Diverse Tasks
        in Human Olfactory Perception preprint
        <https://www.biorxiv.org/content/10.1101/2022.09.01.504602v4>`_.

    .. [2] Benjamin Sanchez-Lengeling, Jennifer N. Wei, Brian K. Lee,
        Richard C. Gerkin, Alán Aspuru-Guzik, Alexander B. Wiltschko
        `Machine Learning for Scent:
        Learning Generalizable Perceptual Representations
        of Small Molecules <https://arxiv.org/abs/1910.10685>`_.

    .. [3] Justin Gilmer, Samuel S. Schoenholz, Patrick F. Riley,
        Oriol Vinyals, George E. Dahl.
        "Neural Message Passing for Quantum Chemistry." ICML 2017.

    Notes
    -----
    This class requires DGL (https://github.com/dmlc/dgl)
    and DGL-LifeSci (https://github.com/awslabs/dgl-lifesci)
    to be installed.
    """

    def __init__(self,
                 n_tasks: int,
                 node_out_feats: int = 64,
                 edge_hidden_feats: int = 128,
                 edge_out_feats: int = 64,
                 num_step_message_passing: int = 3,
                 mpnn_residual: bool = True,
                 message_aggregator_type: str = 'sum',
                 mode: str = 'classification',
                 number_atom_features: int = 134,
                 number_bond_features: int = 6,
                 n_classes: int = 1,
                 nfeat_name: str = 'x',
                 efeat_name: str = 'edge_attr',
                 readout_type: str = 'set2set',
                 num_step_set2set: int = 6,
                 num_layer_set2set: int = 3,
                 ffn_hidden_list: List = [300],
                 ffn_embeddings: int = 256,
                 ffn_activation: str = 'relu',
                 ffn_dropout_p: float = 0.0,
                 ffn_dropout_at_input_no_act: bool = True):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks.
        node_out_feats: int
            The length of the final node representation vectors
            before readout. Default to 64.
        edge_hidden_feats: int
            The length of the hidden edge representation vectors
            for mpnn edge network. Default to 128.
        edge_out_feats: int
            The length of the final edge representation vectors
            before readout. Default to 64.
        num_step_message_passing: int
            The number of rounds of message passing. Default to 3.
        mpnn_residual: bool
            If true, adds residual layer to mpnn layer. Default to True.
        message_aggregator_type: str
            MPNN message aggregator type, 'sum', 'mean' or 'max'.
            Default to 'sum'.
        mode: str
            The model type, 'classification' or 'regression'.
            Default to 'classification'.
        number_atom_features: int
            The length of the initial atom feature vectors. Default to 134.
        number_bond_features: int
            The length of the initial bond feature vectors. Default to 6.
        n_classes: int
            The number of classes to predict per task
            (only used when ``mode`` is 'classification'). Default to 1.
        nfeat_name: str
            For an input graph ``g``, the model assumes that it stores
            node features in ``g.ndata[nfeat_name]`` and will retrieve
            input node features from that. Default to 'x'.
        efeat_name: str
            For an input graph ``g``, the model assumes that it stores
            edge features in ``g.edata[efeat_name]`` and will retrieve
            input edge features from that. Default to 'edge_attr'.
        readout_type: str
            The Readout type, 'set2set' or 'global_sum_pooling'.
            Default to 'set2set'.
        num_step_set2set: int
            Number of steps in set2set readout.
            Used if, readout_type == 'set2set'.
            Default to 6.
        num_layer_set2set: int
            Number of layers in set2set readout.
            Used if, readout_type == 'set2set'.
            Default to 3.
        ffn_hidden_list: List
            List of sizes of hidden layer in the feed-forward network layer.
            Default to [300].
        ffn_embeddings: int
            Size of penultimate layer in the feed-forward network layer.
            This determines the Principal Odor Map dimension.
            Default to 256.
        ffn_activation: str
            Activation function to be used in feed-forward network layer.
            Can choose between 'relu' for ReLU, 'leakyrelu' for LeakyReLU,
            'prelu' for PReLU, 'tanh' for TanH, 'selu' for SELU,
            and 'elu' for ELU.
        ffn_dropout_p: float
            Dropout probability for the feed-forward network layer.
            Default to 0.0
        ffn_dropout_at_input_no_act: bool
            If true, dropout is applied on the input tensor.
            For single layer, it is not passed to an activation function.
        """
        if mode not in ['classification', 'regression']:
            raise ValueError(
                "mode must be either 'classification' or 'regression'")

        super(MPNNPOM, self).__init__()

        self.n_tasks: int = n_tasks
        self.mode: str = mode
        self.n_classes: int = n_classes
        self.nfeat_name: str = nfeat_name
        self.efeat_name: str = efeat_name
        self.readout_type: str = readout_type
        self.ffn_embeddings: int = ffn_embeddings
        self.ffn_activation: str = ffn_activation
        self.ffn_dropout_p: float = ffn_dropout_p

        if mode == 'classification':
            self.ffn_output: int = n_tasks * n_classes
        else:
            self.ffn_output = n_tasks

        self.mpnn: nn.Module = CustomMPNNGNN(
            node_in_feats=number_atom_features,
            node_out_feats=node_out_feats,
            edge_in_feats=number_bond_features,
            edge_hidden_feats=edge_hidden_feats,
            num_step_message_passing=num_step_message_passing,
            residual=mpnn_residual,
            message_aggregator_type=message_aggregator_type)

        self.project_edge_feats: nn.Module = nn.Sequential(
            nn.Linear(number_bond_features, edge_out_feats), nn.ReLU())

        if self.readout_type == 'set2set':
            self.readout_set2set: nn.Module = Set2Set(
                input_dim=node_out_feats + edge_out_feats,
                n_iters=num_step_set2set,
                n_layers=num_layer_set2set)
            ffn_input: int = 2 * (node_out_feats + edge_out_feats)
        elif self.readout_type == 'global_sum_pooling':
            ffn_input = node_out_feats + edge_out_feats
        else:
            raise Exception("readout_type invalid")

        if ffn_embeddings is not None:
            d_hidden_list: List = ffn_hidden_list + [ffn_embeddings]

        self.ffn: nn.Module = CustomPositionwiseFeedForward(
            d_input=ffn_input,
            d_hidden_list=d_hidden_list,
            d_output=self.ffn_output,
            activation=ffn_activation,
            dropout_p=ffn_dropout_p,
            dropout_at_input_no_act=ffn_dropout_at_input_no_act)

    def _readout(self, g: DGLGraph, node_encodings: torch.Tensor,
                 edge_feats: torch.Tensor) -> torch.Tensor:
        """
        Method to execute the readout phase.
        (compute molecules encodings from atom hidden states)

        Readout phase consists of radius 0 combination to fold atom
        and bond embeddings together,
        followed by:
            - a reduce-sum across atoms
                if `self.readout_type == 'global_sum_pooling'`
            - set2set pooling
                if `self.readout_type == 'set2set'`

        Parameters
        ----------
        g: DGLGraph
            A DGLGraph for a batch of graphs.
            It stores the node features in
            ``dgl_graph.ndata[self.nfeat_name]`` and edge features in
            ``dgl_graph.edata[self.efeat_name]``.

        node_encodings: torch.Tensor
            Tensor containing node hidden states.

        edge_feats: torch.Tensor
            Tensor containing edge features.

        Returns
        -------
        batch_mol_hidden_states: torch.Tensor
            Tensor containing batchwise molecule encodings.
        """

        g.ndata['node_emb'] = node_encodings
        g.edata['edge_emb'] = self.project_edge_feats(edge_feats)

        def message_func(edges) -> Dict:
            """
            The message function to generate messages
            along the edges for DGLGraph.send_and_recv()
            """
            src_msg: torch.Tensor = torch.cat(
                (edges.src['node_emb'], edges.data['edge_emb']), dim=1)
            return {'src_msg': src_msg}

        def reduce_func(nodes) -> Dict:
            """
            The reduce function to aggregate the messages
            for DGLGraph.send_and_recv()
            """
            src_msg_sum: torch.Tensor = torch.sum(nodes.mailbox['src_msg'],
                                                  dim=1)
            return {'src_msg_sum': src_msg_sum}

        # radius 0 combination to fold atom and bond embeddings together
        g.send_and_recv(g.edges(),
                        message_func=message_func,
                        reduce_func=reduce_func)

        if self.readout_type == 'set2set':
            batch_mol_hidden_states: torch.Tensor = self.readout_set2set(
                g, g.ndata['src_msg_sum'])
        elif self.readout_type == 'global_sum_pooling':
            batch_mol_hidden_states = dgl.sum_nodes(g, 'src_msg_sum')

        # batch_size x (node_out_feats + edge_out_feats)
        return batch_mol_hidden_states

    def forward(
        self, g: DGLGraph
    ) -> Union[tuple[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]:
        """
        Foward pass for MPNNPOM class. It also returns embeddings for POM.

        Parameters
        ----------
        g: DGLGraph
            A DGLGraph for a batch of graphs. It stores the node features in
            ``dgl_graph.ndata[self.nfeat_name]`` and edge features in
            ``dgl_graph.edata[self.efeat_name]``.

        Returns
        -------
        Union[tuple[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]
            The model output.

        * When self.mode = 'regression',
            its shape will be ``(dgl_graph.batch_size, self.n_tasks)``.
        * When self.mode = 'classification',
            the output consists of probabilities for classes.
            Its shape will be
            ``(dgl_graph.batch_size, self.n_tasks, self.n_classes)``
            if self.n_tasks > 1;
            its shape will be ``(dgl_graph.batch_size, self.n_classes)``
            if self.n_tasks is 1.
        """
        node_feats: torch.Tensor = g.ndata[self.nfeat_name]
        edge_feats: torch.Tensor = g.edata[self.efeat_name]

        node_encodings: torch.Tensor = self.mpnn(g, node_feats, edge_feats)

        molecular_encodings: torch.Tensor = self._readout(
            g, node_encodings, edge_feats)
        if self.readout_type == 'global_sum_pooling':
            molecular_encodings = F.softmax(molecular_encodings, dim=1)

        embeddings: torch.Tensor
        out: torch.Tensor
        embeddings, out = self.ffn(molecular_encodings)

        if self.mode == 'classification':
            if self.n_tasks == 1:
                logits: torch.Tensor = out.view(-1, self.n_classes)
            else:
                logits = out.view(-1, self.n_tasks, self.n_classes)
            proba: torch.Tensor = F.sigmoid(
                logits)  # (batch, n_tasks, classes)
            if self.n_classes == 1:
                proba = proba.squeeze(-1)  # (batch, n_tasks)
            return proba, logits, embeddings
        else:
            return out


class MPNNPOMModel(TorchModel):
    """
    MPNNPOMModel for obtaining a principal odor map
    using multilabel-classification based on the pre-print:
    "A Principal Odor Map Unifies DiverseTasks in Human
        Olfactory Perception" [1]

    * Combine latest node representations and edge features in
        updating node representations, which involves multiple
        rounds of message passing.
    * For each graph, compute its representation by radius 0 combination
        to fold atom and bond embeddings together, followed by
        'set2set' or 'global_sum_pooling' readout.
    * Perform the final prediction using a feed-forward layer.

    References
    ----------
    .. [1] Brian K. Lee, Emily J. Mayhew, Benjamin Sanchez-Lengeling,
        Jennifer N. Wei, Wesley W. Qian, Kelsie Little, Matthew Andres,
        Britney B. Nguyen, Theresa Moloy, Jane K. Parker, Richard C. Gerkin,
        Joel D. Mainland, Alexander B. Wiltschko
        `A Principal Odor Map Unifies Diverse Tasks
        in Human Olfactory Perception preprint
        <https://www.biorxiv.org/content/10.1101/2022.09.01.504602v4>`_.

    .. [2] Benjamin Sanchez-Lengeling, Jennifer N. Wei, Brian K. Lee,
        Richard C. Gerkin, Alán Aspuru-Guzik, Alexander B. Wiltschko
        `Machine Learning for Scent:
        Learning Generalizable Perceptual Representations
        of Small Molecules <https://arxiv.org/abs/1910.10685>`_.

    .. [3] Justin Gilmer, Samuel S. Schoenholz, Patrick F. Riley,
        Oriol Vinyals, George E. Dahl.
        "Neural Message Passing for Quantum Chemistry." ICML 2017.

    Notes
    -----
    This class requires DGL (https://github.com/dmlc/dgl) and DGL-LifeSci
    (https://github.com/awslabs/dgl-lifesci) to be installed.

    The featurizer used with MPNNPOMModel must produce a Deepchem GraphData
    object which should have both 'edge' and 'node' features.
    """

    def __init__(self,
                 n_tasks: int,
                 class_imbalance_ratio: Optional[List] = None,
                 loss_aggr_type: str = 'sum',
                 learning_rate: Union[float, LearningRateSchedule] = 0.001,
                 batch_size: int = 100,
                 node_out_feats: int = 64,
                 edge_hidden_feats: int = 128,
                 edge_out_feats: int = 64,
                 num_step_message_passing: int = 3,
                 mpnn_residual: bool = True,
                 message_aggregator_type: str = 'sum',
                 mode: str = 'regression',
                 number_atom_features: int = 134,
                 number_bond_features: int = 6,
                 n_classes: int = 1,
                 readout_type: str = 'set2set',
                 num_step_set2set: int = 6,
                 num_layer_set2set: int = 3,
                 ffn_hidden_list: List = [300],
                 ffn_embeddings: int = 256,
                 ffn_activation: str = 'relu',
                 ffn_dropout_p: float = 0.0,
                 ffn_dropout_at_input_no_act: bool = True,
                 weight_decay: float = 1e-5,
                 self_loop: bool = False,
                 optimizer_name: str = 'adam',
                 device_name: Optional[str] = None,
                 **kwargs):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks.
        class_imbalance_ratio: Optional[List]
            List of imbalance ratios per task.
        loss_aggr_type: str
            loss aggregation type; 'sum' or 'mean'. Default to 'sum'.
            Only applies to CustomMultiLabelLoss for classification
        learning_rate: Union[float, LearningRateSchedule]
            Learning rate value or scheduler object. Default to 0.001.
        batch_size: int
            Batch size for training. Default to 100.
        node_out_feats: int
            The length of the final node representation vectors
            before readout. Default to 64.
        edge_hidden_feats: int
            The length of the hidden edge representation vectors
            for mpnn edge network. Default to 128.
        edge_out_feats: int
            The length of the final edge representation vectors
            before readout. Default to 64.
        num_step_message_passing: int
            The number of rounds of message passing. Default to 3.
        mpnn_residual: bool
            If true, adds residual layer to mpnn layer. Default to True.
        message_aggregator_type: str
            MPNN message aggregator type, 'sum', 'mean' or 'max'.
            Default to 'sum'.
        mode: str
            The model type, 'classification' or 'regression'.
            Default to 'classification'.
        number_atom_features: int
            The length of the initial atom feature vectors. Default to 134.
        number_bond_features: int
            The length of the initial bond feature vectors. Default to 6.
        n_classes: int
            The number of classes to predict per task
            (only used when ``mode`` is 'classification'). Default to 1.
        readout_type: str
            The Readout type, 'set2set' or 'global_sum_pooling'.
            Default to 'set2set'.
        num_step_set2set: int
            Number of steps in set2set readout.
            Used if, readout_type == 'set2set'.
            Default to 6.
        num_layer_set2set: int
            Number of layers in set2set readout.
            Used if, readout_type == 'set2set'.
            Default to 3.
        ffn_hidden_list: List
            List of sizes of hidden layer in the feed-forward network layer.
            Default to [300].
        ffn_embeddings: int
            Size of penultimate layer in the feed-forward network layer.
            This determines the Principal Odor Map dimension.
            Default to 256.
        ffn_activation: str
            Activation function to be used in feed-forward network layer.
            Can choose between 'relu' for ReLU, 'leakyrelu' for LeakyReLU,
            'prelu' for PReLU, 'tanh' for TanH, 'selu' for SELU,
            and 'elu' for ELU.
        ffn_dropout_p: float
            Dropout probability for the feed-forward network layer.
            Default to 0.0
        ffn_dropout_at_input_no_act: bool
            If true, dropout is applied on the input tensor.
            For single layer, it is not passed to an activation function.
        weight_decay: float
            weight decay value for L1 and L2 regularization. Default to 1e-5.
        self_loop: bool
            Whether to add self loops for the nodes, i.e. edges
            from nodes to themselves. Generally, an MPNNPOMModel
            does not require self loops. Default to False.
        optimizer_name: str
            Name of optimizer to be used from
            [adam, adagrad, adamw, sparseadam, rmsprop, sgd, kfac]
            Default to 'adam'.
        device_name: Optional[str]
            The device on which to run computations. If None, a device is
            chosen automatically.
        kwargs
            This can include any keyword argument of TorchModel.
        """
        model: nn.Module = MPNNPOM(
            n_tasks=n_tasks,
            node_out_feats=node_out_feats,
            edge_hidden_feats=edge_hidden_feats,
            edge_out_feats=edge_out_feats,
            num_step_message_passing=num_step_message_passing,
            mpnn_residual=mpnn_residual,
            message_aggregator_type=message_aggregator_type,
            mode=mode,
            number_atom_features=number_atom_features,
            number_bond_features=number_bond_features,
            n_classes=n_classes,
            readout_type=readout_type,
            num_step_set2set=num_step_set2set,
            num_layer_set2set=num_layer_set2set,
            ffn_hidden_list=ffn_hidden_list,
            ffn_embeddings=ffn_embeddings,
            ffn_activation=ffn_activation,
            ffn_dropout_p=ffn_dropout_p,
            ffn_dropout_at_input_no_act=ffn_dropout_at_input_no_act)

        if class_imbalance_ratio and (len(class_imbalance_ratio) != n_tasks):
            raise Exception("size of class_imbalance_ratio \
                            should be equal to n_tasks")

        if mode == 'regression':
            loss: Loss = L2Loss()
            output_types: List = ['prediction']
        else:
            loss = CustomMultiLabelLoss(
                class_imbalance_ratio=class_imbalance_ratio,
                loss_aggr_type=loss_aggr_type,
                device=device_name)
            output_types = ['prediction', 'loss', 'embedding']

        optimizer: Optimizer = get_optimizer(optimizer_name)
        optimizer.learning_rate = learning_rate
        if device_name is not None:
            device: Optional[torch.device] = torch.device(device_name)
        else:
            device = None
        super(MPNNPOMModel, self).__init__(model,
                                           loss=loss,
                                           output_types=output_types,
                                           optimizer=optimizer,
                                           learning_rate=learning_rate,
                                           batch_size=batch_size,
                                           device=device,
                                           **kwargs)

        self.weight_decay: float = weight_decay
        self._self_loop: bool = self_loop
        self.regularization_loss: Callable = self._regularization_loss

    def _regularization_loss(self) -> torch.Tensor:
        """
        L1 and L2-norm losses for regularization

        Returns
        -------
        torch.Tensor
            sum of l1_norm and l2_norm
        """
        l1_regularization: torch.Tensor = torch.tensor(0., requires_grad=True)
        l2_regularization: torch.Tensor = torch.tensor(0., requires_grad=True)
        for name, param in self.model.named_parameters():
            if 'bias' not in name:
                l1_regularization = l1_regularization + torch.norm(param, p=1)
                l2_regularization = l2_regularization + torch.norm(param, p=2)
        l1_norm: torch.Tensor = self.weight_decay * l1_regularization
        l2_norm: torch.Tensor = self.weight_decay * l2_regularization
        return l1_norm + l2_norm

    def _prepare_batch(
        self, batch: Tuple[List, List, List]
    ) -> Tuple[DGLGraph, List[torch.Tensor], List[torch.Tensor]]:
        """Create batch data for MPNN.

        Parameters
        ----------
        batch: Tuple[List, List, List]
            The tuple is ``(inputs, labels, weights)``.

        Returns
        -------
        g: DGLGraph
            DGLGraph for a batch of graphs.
        labels: list of torch.Tensor or None
            The graph labels.
        weights: list of torch.Tensor or None
            The weights for each sample or
            sample/task pair converted to torch.Tensor.
        """
        inputs: List
        labels: List
        weights: List

        inputs, labels, weights = batch
        dgl_graphs: List[DGLGraph] = [
            graph.to_dgl_graph(self_loop=self._self_loop)
            for graph in inputs[0]
        ]
        g: DGLGraph = dgl.batch(dgl_graphs).to(self.device)
        _, labels, weights = super(MPNNPOMModel, self)._prepare_batch(
            ([], labels, weights))
        return g, labels, weights



import os
import json
import torch
import tempfile
import numpy as np
from tqdm import tqdm
import deepchem as dc
from datetime import datetime
from openpom.models.mpnn_pom import MPNNPOMModel
from openpom.feat.graph_featurizer import GraphFeaturizer, GraphConvConstants
from openpom.utils.data_utils import get_class_imbalance_ratio
from openpom.hyper.configs.model_configs import MPNNPOMConfig
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import logging

logger = logging.getLogger(__name__)
# logging.basicConfig(level=logging.INFO)

DATASET = 'openpom/data/curated_datasets/curated_GS_LF_merged_4983.csv'
SMILES_FIELD = 'nonStereoSMILES'
TASKS = [
    'alcoholic', 'aldehydic', 'alliaceous', 'almond', 'amber', 'animal',
    'anisic', 'apple', 'apricot', 'aromatic', 'balsamic', 'banana', 'beefy',
    'bergamot', 'berry', 'bitter', 'black currant', 'brandy', 'burnt',
    'buttery', 'cabbage', 'camphoreous', 'caramellic', 'cedar', 'celery',
    'chamomile', 'cheesy', 'cherry', 'chocolate', 'cinnamon', 'citrus',
    'clean', 'clove', 'cocoa', 'coconut', 'coffee', 'cognac', 'cooked',
    'cooling', 'cortex', 'coumarinic', 'creamy', 'cucumber', 'dairy', 'dry',
    'earthy', 'ethereal', 'fatty', 'fermented', 'fishy', 'floral', 'fresh',
    'fruit skin', 'fruity', 'garlic', 'gassy', 'geranium', 'grape',
    'grapefruit', 'grassy', 'green', 'hawthorn', 'hay', 'hazelnut', 'herbal',
    'honey', 'hyacinth', 'jasmin', 'juicy', 'ketonic', 'lactonic', 'lavender',
    'leafy', 'leathery', 'lemon', 'lily', 'malty', 'meaty', 'medicinal',
    'melon', 'metallic', 'milky', 'mint', 'muguet', 'mushroom', 'musk',
    'musty', 'natural', 'nutty', 'odorless', 'oily', 'onion', 'orange',
    'orangeflower', 'orris', 'ozone', 'peach', 'pear', 'phenolic', 'pine',
    'pineapple', 'plum', 'popcorn', 'potato', 'powdery', 'pungent', 'radish',
    'raspberry', 'ripe', 'roasted', 'rose', 'rummy', 'sandalwood', 'savory',
    'sharp', 'smoky', 'soapy', 'solvent', 'sour', 'spicy', 'strawberry',
    'sulfurous', 'sweaty', 'sweet', 'tea', 'terpenic', 'tobacco', 'tomato',
    'tropical', 'vanilla', 'vegetable', 'vetiver', 'violet', 'warm', 'waxy',
    'weedy', 'winey', 'woody'
]


def save_checkpoint(self,
                    max_checkpoints_to_keep: int = 5,
                    model_dir=None,
                    ckpt_name: str = 'best_checkpoint') -> None:
    """
        Save a checkpoint to disk.

        Parameters
        ----------
        max_checkpoints_to_keep: int
            the maximum number of checkpoints to keep.  Older checkpoints are discarded.
        model_dir: str, default None
            Model directory to save checkpoint to. If None, revert to self.model_dir
        """
    self._ensure_built()
    if model_dir is None:
        model_dir = self.model_dir
    if not os.path.exists(model_dir):
        os.makedirs(model_dir)

    # Save the checkpoint to a file.

    data = {
        'model_state_dict': self.model.state_dict(),
        'optimizer_state_dict': self._pytorch_optimizer.state_dict(),
        'global_step': self._global_step
    }
    temp_file = os.path.join(model_dir, f'{ckpt_name}.pt')
    torch.save(data, temp_file)

    # Rename and delete older files.
    paths = [
        os.path.join(model_dir, f'{ckpt_name}%d.pt' % (i + 1))
        for i in range(max_checkpoints_to_keep)
    ]
    if os.path.exists(paths[-1]):
        os.remove(paths[-1])
    for i in reversed(range(max_checkpoints_to_keep - 1)):
        if os.path.exists(paths[i]):
            os.rename(paths[i], paths[i + 1])
    os.rename(temp_file, paths[0])


class CV:
    """
    K-FOLD CROSS VALIDATION for MPNNPOM model
    with custom stratification splitting
    """

    def __init__(self, model_builder, n_folds, device=None) -> None:
        self.model_builder = model_builder
        self.n_folds = n_folds
        self.device = device

    def _deepchem_splitter(self, dataset):
        randomstratifiedsplitter = dc.splits.RandomStratifiedSplitter()
        return randomstratifiedsplitter.k_fold_split(dataset=dataset,
                                                     k=self.n_folds)

    def generate_folds(self, dataset, splitter='deepchem'):
        if splitter == 'deepchem':
            self.folds_list = self._deepchem_splitter(dataset)
        elif splitter == 'skmultilearn':
            raise NotImplementedError
        return self.folds_list

    def cross_validation(self,
                         model_params,
                         logdir=None,
                         max_epoch=100,
                         metric=None,
                         save_best_ckpt=False):
        logger.info("hyperparameters: %s" % str(model_params))
        all_folds_train_scores = []
        all_folds_val_scores = []
        if metric is None:
            metric = dc.metrics.Metric(dc.metrics.roc_auc_score)
        for fold_num, (train_dataset,
                       valid_dataset) in enumerate(self.folds_list):
            logger.info("Fitting model %d/%d folds" %
                        (fold_num + 1, self.n_folds))
            folder_name = f"fold_{fold_num + 1}_trial_count_{model_params['trial_count']}_{str(datetime.now())}"

            if logdir is not None:
                model_dir = os.path.join(logdir, folder_name)
                logger.info("model_dir is %s" % model_dir)
                try:
                    os.makedirs(model_dir)
                except OSError:
                    if not os.path.isdir(model_dir):
                        logger.info(
                            "Error creating model_dir, using tempfile directory"
                        )
                        model_dir = tempfile.mkdtemp()
            else:
                model_dir = tempfile.mkdtemp()

            model_params['model_dir'] = model_dir
            model_params['class_imbalance_ratio'] = get_class_imbalance_ratio(
                train_dataset)
            if self.device is not None:
                model_params['device_name'] = self.device
            model = self.model_builder(**model_params)

            best_train_score = 0  # train score for best validation
            best_val_score = 0
            error = ""
            try:
                for epoch in tqdm(range(1, max_epoch + 1)):
                    loss = model.fit(train_dataset,
                                     nb_epoch=1,
                                     max_checkpoints_to_keep=1,
                                     deterministic=False,
                                     restore=epoch > 1)

                    train_scores = model.evaluate(train_dataset,
                                                  [metric])['roc_auc_score']
                    valid_scores = model.evaluate(valid_dataset,
                                                  [metric])['roc_auc_score']
                    if valid_scores > best_val_score:
                        best_val_score = valid_scores
                        best_train_score = train_scores
                        if save_best_ckpt:
                            save_checkpoint(model, 1, None,
                                            f'best_ckpt_{fold_num}_')
                    logger.info(
                        f"epoch {epoch}/{max_epoch} ; loss = {loss}; train_scores = {train_scores}; test_scores = {valid_scores}"
                    )
            except Exception as e:
                error = f"Training error: {e}"

            all_folds_train_scores.append(best_train_score)
            all_folds_val_scores.append(best_val_score)

            try:
                os.remove(os.path.join(model_dir, 'checkpoint1.pt'))
            except:
                pass
            del model
            torch.cuda.empty_cache()

        mean_train_score = np.asarray(all_folds_train_scores).mean()
        mean_val_score = np.asarray(all_folds_val_scores).mean()
        logger.info("Results:")
        logger.info(f"hyperparameters: {str(model_params)}")
        logger.info(f"fold train scores: {all_folds_train_scores}")
        logger.info(f"fold validation scores: {all_folds_val_scores}")
        logger.info(f"mean train score: {mean_train_score}")
        logger.info(f"mean validation score: {mean_val_score}")
        return mean_train_score, mean_val_score, error


def random_search_cv(tasks=TASKS,
                     dataset=DATASET,
                     smiles_field=SMILES_FIELD,
                     n_folds=2,
                     n_trials=1,
                     logdir='./models',
                     max_epoch=10,
                     save_best_ckpt=False):
    # get dataset
    featurizer = GraphFeaturizer()
    smiles_field = smiles_field
    loader = dc.data.CSVLoader(tasks=tasks,
                               feature_field=smiles_field,
                               featurizer=featurizer)
    input_file = dataset
    dataset = loader.create_dataset(inputs=[input_file])
    n_tasks = len(dataset.tasks)

    n_folds = n_folds
    n_trials = n_trials
    logdir = logdir
    max_epoch = max_epoch
    device = 'cuda' if torch.cuda.is_available() else 'cpu'

    # Metric
    metric = dc.metrics.Metric(dc.metrics.roc_auc_score)

    model_builder = lambda **params: MPNNPOMModel(
        n_tasks=n_tasks,
        mode='classification',
        number_atom_features=GraphConvConstants.ATOM_FDIM,
        number_bond_features=GraphConvConstants.BOND_FDIM,
        n_classes=1,
        **params)

    cv = CV(model_builder=model_builder, n_folds=n_folds, device=device)
    cv.generate_folds(dataset=dataset, splitter='deepchem')

    try:
        file_name = f"{n_trials}_trials_params.json"
        file_path = os.path.join('./examples/trials', file_name)

        with open(file_path, 'r') as json_file:
            trials_dict = json.load(json_file)
    except:
        trials_dict, _ = MPNNPOMConfig.generate_hyperparams_random(
            n_trials=n_trials, dir='./examples/trials')

    logger.info("Starting random search crosss validation:")
    best_train_score = 0
    best_validation_score = 0
    best_hyperparams = {}
    all_scores = {}
    for trial_count, model_params in tqdm(trials_dict.items()):
        logger.info(f"{trial_count} starting:")
        model_params['trial_count'] = trial_count
        trial_start_time = datetime.now()
        mean_train_score, mean_val_score, error = cv.cross_validation(
            model_params=model_params,
            logdir=logdir,
            max_epoch=max_epoch,
            save_best_ckpt=save_best_ckpt,
            metric=metric)
        trial_end_time = datetime.now()

        all_scores[trial_count] = {
            'mean_train_score': mean_train_score,
            'mean_val_score': mean_val_score
        }

        current_date_time = str(datetime.now())
        log_file = os.path.join(
            logdir, f'results_{trial_count}_trial_{current_date_time}.txt')
        with open(log_file, 'w+') as f:
            f.write("Hyperparameters dictionary %s\n" % str(model_params))
            f.write("validation score %f\n" % mean_val_score)
            f.write("train_score: %f\n" % mean_train_score)
            f.write("trial_time: %s\n" %
                    str(trial_end_time - trial_start_time))
            f.write("error: %s\n" % error)

        if mean_val_score > best_validation_score:
            best_train_score = mean_train_score
            best_validation_score = mean_val_score
            best_hyperparams = model_params

    logger.info("Best hyperparameters: %s" % str(best_hyperparams))
    logger.info("best train_score: %f" % best_train_score)
    logger.info("best validation_score: %f" % best_validation_score)

    current_date_time = str(datetime.now())
    log_file = os.path.join(
        logdir, f'results_{n_trials}_trials_{current_date_time}.txt')

    with open(log_file, 'w+') as f:
        f.write("Best Hyperparameters dictionary %s\n" % str(best_hyperparams))
        f.write("Best validation score %f\n" % best_validation_score)
        f.write("Best train_score: %f\n" % best_train_score)


if __name__ == "__main__":
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-f",
                        "--n_folds",
                        default=2,
                        type=int,
                        help="Number of folds for cross-validation")
    parser.add_argument("-t",
                        "--n_trials",
                        default=2,
                        type=int,
                        help="Number of trials for random search cv")
    parser.add_argument("-d",
                        "--logdir",
                        default="./models",
                        help="Log directory to store results")
    parser.add_argument("-e",
                        "--max_epoch",
                        default=10,
                        type=int,
                        help="Number of epochs for each fold per trial")
    parser.add_argument("-c",
                        "--save_best_ckpt",
                        action="store_true",
                        help="Whether to save best checkpoints?")
    args = vars(parser.parse_args())

    n_folds = args['n_folds']
    n_trials = args['n_trials']
    logdir = args['logdir']
    max_epoch = args['max_epoch']
    save_best_ckpt = args['save_best_ckpt']
    random_search_cv(n_folds=n_folds,
                     n_trials=n_trials,
                     logdir=logdir,
                     max_epoch=max_epoch,
                     save_best_ckpt=save_best_ckpt)



import torch
from openpom.feat.graph_featurizer import GraphFeaturizer
from openpom.layers.pom_ffn import CustomPositionwiseFeedForward
from openpom.layers.pom_mpnn_gnn import CustomMPNNGNN


def test_custom_position_wise_feed_forward():
    """Test invoking CustomPositionwiseFeedForward."""
    torch.manual_seed(0)
    input_ar = torch.tensor([[1., 2.], [5., 6.]])
    ffn = CustomPositionwiseFeedForward(d_input=2,
                                        d_hidden_list=[16, 3],
                                        d_output=2,
                                        activation='leakyrelu',
                                        dropout_p=0.1,
                                        dropout_at_input_no_act=True,
                                        batch_norm=True)
    assert len(ffn.batchnorms) == 2
    embbedding_result, output_result = ffn(input_ar)
    assert embbedding_result.shape == (2, 3)
    assert output_result.shape == (2, 2)


def test_custom_position_wise_feed_forward_no_batchnorm():
    """Test invoking CustomPositionwiseFeedForward without batch norm"""
    torch.manual_seed(0)
    input_ar = torch.tensor([[1., 2.], [5., 6.]])
    ffn = CustomPositionwiseFeedForward(d_input=2,
                                        d_hidden_list=[16, 3],
                                        d_output=2,
                                        activation='leakyrelu',
                                        dropout_p=0.1,
                                        dropout_at_input_no_act=True,
                                        batch_norm=False)
    assert not hasattr(ffn, 'batchnorms')
    embbedding_result, output_result = ffn(input_ar)
    assert embbedding_result.shape == (2, 3)
    assert output_result.shape == (2, 2)


def test_custom_mpnn_gnn_residual_sum():
    """
    Test invoking CustomMPNNGNN with residual
    and message summation
    """
    torch.manual_seed(0)
    mpnngnn = CustomMPNNGNN(node_in_feats=134,
                            edge_in_feats=6,
                            node_out_feats=4,
                            edge_hidden_feats=10,
                            num_step_message_passing=3,
                            residual=True,
                            message_aggregator_type='sum')
    assert mpnngnn.gnn_layer.res_fc is not None
    assert mpnngnn.gnn_layer.reducer.__name__ == 'sum'

    featurizer = GraphFeaturizer()
    graph = featurizer.featurize('O=C=O')[0]
    g = graph.to_dgl_graph(self_loop=False)

    node_feats = g.ndata['x']
    edge_feats = g.edata['edge_attr']

    node_encodings = mpnngnn(g, node_feats, edge_feats)
    assert node_encodings.shape == (3, 4)


def test_custom_mpnn_gnn_no_residual_mean():
    """
    Test invoking CustomMPNNGNN with no residual
    and message mean aggregation
    """
    torch.manual_seed(0)
    mpnngnn1 = CustomMPNNGNN(node_in_feats=134,
                             edge_in_feats=6,
                             node_out_feats=10,
                             edge_hidden_feats=10,
                             num_step_message_passing=3,
                             residual=False,
                             message_aggregator_type='mean')
    assert mpnngnn1.gnn_layer.res_fc is None
    assert mpnngnn1.gnn_layer.reducer.__name__ == 'mean'

    featurizer = GraphFeaturizer()
    graph = featurizer.featurize('O=C=O')[0]
    g = graph.to_dgl_graph(self_loop=False)

    node_feats = g.ndata['x']
    edge_feats = g.edata['edge_attr']

    node_encodings1 = mpnngnn1(g, node_feats, edge_feats)
    assert node_encodings1.shape == (3, 10)



import torch.nn as nn
from dgl.nn.pytorch import NNConv
from dgllife.model.gnn import MPNNGNN


class CustomMPNNGNN(MPNNGNN):
    """
    Customized MPNNGNN layer based MPNNGNN layer in dgllife library.

    Additional options:
    -> toggle for residual in gnn layer
    -> choice for message aggregator type

    MPNN is introduced in `Neural Message Passing for Quantum Chemistry
    <https://arxiv.org/abs/1704.01212>`__.

    This class performs message passing in MPNN
    and returns the updated node representations.
    """

    def __init__(self,
                 node_in_feats: int = 50,
                 edge_in_feats: int = 50,
                 node_out_feats: int = 64,
                 edge_hidden_feats: int = 128,
                 num_step_message_passing: int = 6,
                 residual: bool = True,
                 message_aggregator_type: str = 'sum'):
        """
        Parameters
        ----------
        node_in_feats: int
            Size for the input node features.
        node_out_feats: int
            Size for the output node representations. Default to 64.
        edge_in_feats: int
            Size for the input edge features. Default to 128.
        edge_hidden_feats: int
            Size for the hidden edge representations.
        num_step_message_passing: int
            Number of message passing steps. Default to 6.
        residual: bool
            If true, adds residual layer to gnn layer
        message_aggregator_type: str
            message aggregator type, 'sum', 'mean' or 'max'
        """
        super(CustomMPNNGNN,
              self).__init__(node_in_feats=node_in_feats,
                             edge_in_feats=edge_in_feats,
                             node_out_feats=node_out_feats,
                             edge_hidden_feats=edge_hidden_feats,
                             num_step_message_passing=num_step_message_passing)

        edge_network = nn.Sequential(
            nn.Linear(edge_in_feats, edge_hidden_feats), nn.ReLU(),
            nn.Linear(edge_hidden_feats, node_out_feats * node_out_feats))
        self.gnn_layer = NNConv(in_feats=node_out_feats,
                                out_feats=node_out_feats,
                                edge_func=edge_network,
                                aggregator_type=message_aggregator_type,
                                residual=residual)



from openpom.utils.optimizer import get_optimizer


def test_get_optimizer():
    """
    Test get_optimizer function
    """
    optim = get_optimizer(optimizer_name='adam')
    assert optim.__class__.__name__ == 'Adam'

    optim = get_optimizer(optimizer_name='adagrad')
    assert optim.__class__.__name__ == 'AdaGrad'

    optim = get_optimizer(optimizer_name='adamw')
    assert optim.__class__.__name__ == 'AdamW'

    optim = get_optimizer(optimizer_name='sparseadam')
    assert optim.__class__.__name__ == 'SparseAdam'

    optim = get_optimizer(optimizer_name='rmsprop')
    assert optim.__class__.__name__ == 'RMSProp'

    optim = get_optimizer(optimizer_name='sgd')
    assert optim.__class__.__name__ == 'GradientDescent'

    optim = get_optimizer(optimizer_name='kfac')
    assert optim.__class__.__name__ == 'KFAC'

    optim = get_optimizer(optimizer_name='APPLE')  # test invalid name
    assert optim.__class__.__name__ == 'Adam'



import numpy as np
from rdkit import Chem
from typing import List, Union, Dict, Sequence
from deepchem.utils.typing import RDKitAtom, RDKitBond, RDKitMol
from deepchem.feat.base_classes import MolecularFeaturizer
from deepchem.feat.graph_data import GraphData
from deepchem.utils.molecule_feature_utils import get_atom_total_degree_one_hot
from deepchem.utils.molecule_feature_utils \
    import get_atom_formal_charge_one_hot
from deepchem.utils.molecule_feature_utils \
    import get_atom_total_num_Hs_one_hot
from deepchem.utils.molecule_feature_utils \
    import get_atom_hybridization_one_hot
from openpom.utils.molecule_feature_utils \
    import get_atomic_num_one_hot, get_atom_total_valence_one_hot
import logging

logger = logging.getLogger(__name__)


class GraphConvConstants(object):
    """
    A class for holding featurization parameters.
    """

    MAX_ATOMIC_NUM = 100
    ATOM_FEATURES: Dict[str, List[int]] = {
        'valence': [0, 1, 2, 3, 4, 5, 6],
        'degree': [0, 1, 2, 3, 4, 5],
        'num_Hs': [0, 1, 2, 3, 4],
        'formal_charge': [-1, -2, 1, 2, 0],
        'atomic_num': list(range(MAX_ATOMIC_NUM)),
    }
    ATOM_FEATURES_HYBRIDIZATION: List[str] = [
        "SP", "SP2", "SP3", "SP3D", "SP3D2"
    ]
    # Dimension of atom feature vector
    ATOM_FDIM = sum(len(choices) + 1
                    for choices in ATOM_FEATURES.values()) + len(
                        ATOM_FEATURES_HYBRIDIZATION) + 1
    # len(choices) +1 and len(ATOM_FEATURES_HYBRIDIZATION)
    # + 1 to include room for unknown set
    BOND_FDIM = 6


def atom_features(atom: RDKitAtom) -> Sequence[Union[bool, int, float]]:
    """
    Helper method used to compute atom feature vector.

    Parameters
    ----------
    atom: RDKitAtom
        Atom to compute features on.

    Returns
    -------
    features: Sequence[Union[bool, int, float]]
        A list of atom features.
    """
    if atom is None:
        features: Sequence[Union[bool, int,
                                 float]] = [0] * GraphConvConstants.ATOM_FDIM

    else:
        features = []
        features += get_atom_total_valence_one_hot(
            atom, GraphConvConstants.ATOM_FEATURES['valence'])
        features += get_atom_total_degree_one_hot(
            atom, GraphConvConstants.ATOM_FEATURES['degree'])
        features += get_atom_total_num_Hs_one_hot(
            atom, GraphConvConstants.ATOM_FEATURES['num_Hs'])
        features += get_atom_formal_charge_one_hot(
            atom, GraphConvConstants.ATOM_FEATURES['formal_charge'])
        features += get_atomic_num_one_hot(
            atom, GraphConvConstants.ATOM_FEATURES['atomic_num'])
        features += get_atom_hybridization_one_hot(
            atom, GraphConvConstants.ATOM_FEATURES_HYBRIDIZATION, True)
        features = [int(feature) for feature in features]
    return features


def bond_features(bond: RDKitBond) -> Sequence[Union[bool, int, float]]:
    """
    Helper method used to compute bond feature vector.

    Parameters
    ----------
    bond: RDKitBond
        Bond to compute features on.

    Returns
    -------
    features: Sequence[Union[bool, int, float]]
        A list of bond features.
    """
    if bond is None:
        b_features: Sequence[Union[
            bool, int, float]] = [1] + [0] * (GraphConvConstants.BOND_FDIM - 1)

    else:
        bt = bond.GetBondType()
        b_features = [
            0, bt == Chem.rdchem.BondType.SINGLE,
            bt == Chem.rdchem.BondType.DOUBLE,
            bt == Chem.rdchem.BondType.TRIPLE,
            bt == Chem.rdchem.BondType.AROMATIC,
            bond.IsInRing()
        ]

    return b_features


class GraphFeaturizer(MolecularFeaturizer):
    """
    This class is a featurizer for GNN (MESSAGE PASSING) implementation for
    Principal Odor Map.

    The default node(atom) and edge(bond) representations are based on
    `A Principal Odor Map Unifies Diverse Tasks in Human Olfactory Perception
    preprint <https://www.biorxiv.org/content/10.1101/2022.09.01.504602v4>`_.

    The default node representation are constructed by concatenating
    the following values, and the feature length is 134.

    - Valence: A one-hot vector for total valence (0-6) of an atom.
    - Degree: A one-hot vector of the degree (0-5) of this atom.
    - Number of Hydrogens: A one-hot vector of the number of hydrogens
      (0-4) that this atom connected.
    - Formal charge: Integer electronic charge, -1, -2, 1, 2, 0.
    - Atomic num: A one-hot vector of this atom, in a range of first 100 atoms.
    - Hybridization: A one-hot vector of "SP", "SP2", "SP3", "SP3D", "SP3D2".

    The default edge representation are constructed by concatenating
    the following values, and the feature length is 6.

    - Bond type: A one-hot vector of the bond type,
      "single", "double", "triple", or "aromatic".
    - Is in ring: Boolean value to specify whether
      the bond is in a ring or not.

    If you want to know more details about features,
    please check the paper [1]_ and utilities in
    deepchem.utils.molecule_feature_utils.py.

    References
    ----------
    .. [1] Kearnes, Steven, et al.
       "Molecular graph convolutions: moving beyond fingerprints."
        Journal of computer-aided molecular design 30.8 (2016):595-608.

    Note
    ----
    This class requires RDKit to be installed.

    """

    def __init__(self, is_adding_hs=False):
        """
        Parameters
        ----------
        is_adding_hs: bool, default False
            Whether to add Hs or not.
        """
        self.is_adding_hs = is_adding_hs
        super(GraphFeaturizer).__init__()

    def _construct_bond_index(self, datapoint: RDKitMol) -> np.ndarray:
        """
        Construct edge (bond) index

        Parameters
        ----------
        datapoint: RDKitMol
            RDKit mol object.

        Returns
        -------
        edge_index: np.ndarray
            Edge (Bond) index

        """
        src: List[int] = []
        dest: List[int] = []
        for bond in datapoint.GetBonds():
            start, end = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
            src += [start, end]
            dest += [end, start]
        return np.asarray([src, dest], dtype=int)

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> GraphData:
        """Calculate molecule graph features from RDKit mol object.

        Parameters
        ----------
        datapoint: RDKitMol
            RDKit mol object.

        Returns
        -------
        graph: GraphData
            A molecule graph object with features:
            - node_features: Node feature matrix with shape
              [num_nodes, num_node_features]
            - edge_index: Graph connectivity in COO format with shape
              [2, num_edges]
            - edge_features: Edge feature matrix with shape
              [num_edges, num_edge_features]
        """
        if isinstance(datapoint, Chem.rdchem.Mol):
            if self.is_adding_hs:
                datapoint = Chem.AddHs(datapoint)
        else:
            raise ValueError(
                "Feature field should contain smiles for featurizer!")

        # get atom features
        f_atoms: np.ndarray = np.asarray(
            [atom_features(atom) for atom in datapoint.GetAtoms()],
            dtype=float)

        # get edge(bond) features
        if len(datapoint.GetBonds()) == 0:
            f_bonds: np.ndarray = np.empty((0, GraphConvConstants.BOND_FDIM))
        else:
            f_bonds_list = []
            for bond in datapoint.GetBonds():
                b_feat = 2 * [bond_features(bond)]
                f_bonds_list.extend(b_feat)
            f_bonds = np.asarray(f_bonds_list, dtype=float)

        # get edge index
        edge_index: np.ndarray = self._construct_bond_index(datapoint)

        return GraphData(node_features=f_atoms,
                         edge_index=edge_index,
                         edge_features=f_bonds)



import torch
import tempfile
import numpy as np
import deepchem as dc
from deepchem.data.data_loader import CSVLoader
from openpom.feat.graph_featurizer import GraphFeaturizer
from openpom.models.mpnn_pom import MPNNPOM, MPNNPOMModel
from openpom.utils.data_utils import get_class_imbalance_ratio


def test_mpnnpom_model_classification():
    """
    Test MPNNPOMModel class for classification
    """
    torch.manual_seed(0)
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    torch.set_default_device(device)

    # load sample dataset
    featurizer = GraphFeaturizer()
    smiles_field = 'smiles'
    tasks = ['fruity', 'green', 'herbal', 'sweet', 'woody']
    loader = CSVLoader(tasks=tasks,
                       feature_field=smiles_field,
                       featurizer=featurizer)
    input_file = \
        'openpom/models/test/assets/test_dataset_sample_7.csv'
    dataset = loader.create_dataset(inputs=[input_file])
    class_imbalance_ratio = get_class_imbalance_ratio(dataset=dataset)

    model = MPNNPOMModel(n_tasks=len(tasks),
                         batch_size=2,
                         class_imbalance_ratio=class_imbalance_ratio,
                         mode="classification",
                         n_classes=1,
                         device_name=device)

    assert isinstance(model.model, MPNNPOM)
    # overfit test
    model.fit(dataset, nb_epoch=50)
    metric = dc.metrics.Metric(dc.metrics.roc_auc_score)
    scores = model.evaluate(dataset, [metric])
    assert scores['roc_auc_score'] > 0.9


def test_mpnnpom_model_regression():
    """
    Test MPNNPOMModel class for regression
    """
    torch.manual_seed(0)
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    torch.set_default_device(device)

    # load sample dataset
    featurizer = GraphFeaturizer()
    smiles_field = 'smiles'
    tasks = ['y']
    loader = CSVLoader(tasks=tasks,
                       feature_field=smiles_field,
                       featurizer=featurizer)
    input_file = \
        'openpom/models/test/assets/test_regression_sample.csv'
    dataset = loader.create_dataset(inputs=[input_file])

    model = MPNNPOMModel(
        n_tasks=len(tasks),
        batch_size=2,
        mode="regression",
    )

    assert isinstance(model.model, MPNNPOM)
    # overfit test
    model.fit(dataset, nb_epoch=100)
    metric = dc.metrics.Metric(dc.metrics.mean_absolute_error,
                               mode="regression")
    scores = model.evaluate(dataset, [metric])
    assert scores['mean_absolute_error'] < 0.5


def test_mpnnpom_model_reload():
    """
    Test MPNNPOMModel class for model reload
    """
    torch.manual_seed(0)
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    torch.set_default_device(device)

    # load sample dataset
    featurizer = GraphFeaturizer()
    smiles_field = 'smiles'
    tasks = ['y']
    loader = CSVLoader(tasks=tasks,
                       feature_field=smiles_field,
                       featurizer=featurizer)
    input_file = \
        'openpom/models/test/assets/test_regression_sample.csv'
    dataset = loader.create_dataset(inputs=[input_file])

    # initialize the model
    model_dir = tempfile.mkdtemp()
    model = MPNNPOMModel(n_tasks=len(tasks),
                         batch_size=2,
                         mode="regression",
                         model_dir=model_dir)

    # fit the model
    model.fit(dataset, nb_epoch=10)

    # reload the model
    reloaded_model = MPNNPOMModel(n_tasks=len(tasks),
                                  batch_size=2,
                                  mode="regression",
                                  model_dir=model_dir)
    reloaded_model.restore()

    orig_predict = model.predict(dataset)
    reloaded_predict = reloaded_model.predict(dataset)
    assert np.all(orig_predict == reloaded_predict)



import torch
from openpom.utils.loss import CustomMultiLabelLoss


def test_custom_multilabel_loss_sum():
    """
    Test CustomMultiLabelLoss with class imbalance ratio and sum aggregation
    """
    class_imbalance_ratio = [1.0, 0.5, 0.25]
    loss = CustomMultiLabelLoss(class_imbalance_ratio, loss_aggr_type='sum')
    loss_fn = loss._create_pytorch_loss()
    sample_output = torch.Tensor([[[0.75], [0.25],
                                   [0.90]]])  # shape: (1, 3, 1)
    sample_target = torch.Tensor([[1.0, 0.0, 0.0]])

    computed_loss = loss_fn(sample_output, sample_target)
    assert computed_loss.shape == (1, 3)
    assert torch.allclose(computed_loss,
                          torch.Tensor([[0.7822, 0.7822, 0.7822]]),
                          atol=0.001)


def test_custom_multilabel_loss_mean():
    """
    Test CustomMultiLabelLoss with class imbalance ratio and mean aggregation
    """
    class_imbalance_ratio = [1.0, 0.5, 0.25]
    loss = CustomMultiLabelLoss(class_imbalance_ratio, loss_aggr_type='mean')
    loss_fn = loss._create_pytorch_loss()
    sample_output = torch.Tensor([[[0.75], [0.25],
                                   [0.90]]])  # shape: (1, 3, 1)
    sample_target = torch.Tensor([[[1.0], [0.0], [0.0]]])

    computed_loss = loss_fn(sample_output, sample_target)
    assert computed_loss.shape == (1, 3)
    assert torch.allclose(computed_loss,
                          torch.Tensor([[0.2607, 0.2607, 0.2607]]),
                          atol=0.001)


def test_custom_multilabel_no_class_imbalance():
    """
    Test CustomMultiLabelLoss with no class imbalance ratio
    """
    loss = CustomMultiLabelLoss(loss_aggr_type='mean')
    loss_fn = loss._create_pytorch_loss()
    sample_output = torch.Tensor([[[0.75], [0.25],
                                   [0.90]]])  # shape: (1, 3, 1)
    sample_target = torch.Tensor([[[1.0], [0.0], [0.0]]])

    computed_loss = loss_fn(sample_output, sample_target)
    assert computed_loss.shape == (1, 3)
    assert torch.allclose(computed_loss,
                          torch.Tensor([[0.7064, 0.7064, 0.7064]]),
                          atol=0.001)



from openpom.utils.molecule_feature_utils \
    import get_atomic_num_one_hot
from openpom.utils.molecule_feature_utils \
    import get_atom_total_valence_one_hot
from openpom.feat.graph_featurizer import GraphConvConstants
from rdkit import Chem
import numpy as np


def test_atomic_num_one_hot():
    """
    Test for get_atomic_num_one_hot feat util
    """
    smiles = 'C'
    m = Chem.MolFromSmiles(smiles)
    atom = m.GetAtoms()[0]
    f_atomic = get_atomic_num_one_hot(
        atom, GraphConvConstants.ATOM_FEATURES['atomic_num'])
    req_f = list(
        np.zeros((len(GraphConvConstants.ATOM_FEATURES['atomic_num']) + 1, ),
                 dtype=float))
    req_f[5] = 1.0
    assert len(f_atomic) == len(req_f)
    assert f_atomic == req_f


def test_total_valence_one_hot():
    """
    Test for get_atom_total_valence_one_hot feat util
    """
    smiles = 'C'
    m = Chem.MolFromSmiles(smiles)
    atom = m.GetAtoms()[0]
    f_valence = get_atom_total_valence_one_hot(
        atom, GraphConvConstants.ATOM_FEATURES['valence'])
    req_f = list(
        np.zeros((len(GraphConvConstants.ATOM_FEATURES['valence']) + 1, ),
                 dtype=float))
    req_f[4] = 1.0
    assert len(f_valence) == len(req_f)
    assert f_valence == req_f



import pandas as pd
import numpy as np
from typing import List
from deepchem.data.datasets import DiskDataset


def get_class_imbalance_ratio(dataset: DiskDataset) -> List:
    """
    Get imbalance ratio per task from DiskDataset

    Imbalance ratio per label (IRLbl): Let M be an MLD with a set of
    labels L and Yi be the label-set of the ith instance. IRLbl is calcu-
    lated for the label λ as the ratio between the majority label and
    the label λ, where IRLbl is 1 for the most frequent label and a
    greater value for the rest. The larger the value of IRLbl, the higher
    the imbalance level for the concerned label.

    Parameters
    ---------
    dataset: DiskDataset
        Deepchem diskdataset object to get class imbalance ratio

    Returns
    -------
    class_imbalance_ratio: List
        List of imbalance ratios per task

    References
    ----------
    .. TarekegnA.N. et al.
       "A review of methods for imbalanced multi-label classification"
       Pattern Recognit. (2021)
    """
    if not isinstance(dataset, DiskDataset):
        raise Exception("The dataset should be a deepchem DiskDataset")
    df: pd.DataFrame = pd.DataFrame(dataset.y)
    class_counts: np.ndarray = df.sum().to_numpy()
    max_count: int = max(class_counts)
    class_imbalance_ratio: List = (class_counts / max_count).tolist()
    return class_imbalance_ratio



from typing import List
from deepchem.utils.typing import RDKitAtom
from deepchem.utils.molecule_feature_utils import one_hot_encode


def get_atomic_num_one_hot(atom: RDKitAtom,
                           allowable_set: List[int],
                           include_unknown_set: bool = True) -> List[float]:
    """
    Get a one-hot feature about atomic number of the given atom.

    Parameters
    ---------
    atom: RDKitAtom
        RDKit atom object
    allowable_set: List[int]
        The range of atomic numbers to consider.
    include_unknown_set: bool, default False
        If true, the index of all types not in
        `allowable_set` is `len(allowable_set)`.

    Returns
    -------
    List[float]
        A one-hot vector of atomic number of the given atom.
        If `include_unknown_set` is False, the length is
        `len(allowable_set)`.
        If `include_unknown_set` is True, the length is
        `len(allowable_set) + 1`.

    """
    return one_hot_encode(atom.GetAtomicNum() - 1, allowable_set,
                          include_unknown_set)


def get_atom_total_valence_one_hot(
        atom: RDKitAtom,
        allowable_set: List[int],
        include_unknown_set: bool = True) -> List[float]:
    """Get a one-hot feature for total valence of an atom.

    Parameters
    ---------
    atom: rdkit.Chem.rdchem.Atom
        RDKit atom object
    allowable_set: List[int]
        Atom total valence to consider.
    include_unknown_set: bool, default True
        If true, the index of all types not in
        `allowable_set` is `len(allowable_set)`.

    Returns
    -------
    List[float]
        A one-hot vector for total valence an atom has.
        If `include_unknown_set` is False, the length is
        `len(allowable_set)`.
        If `include_unknown_set` is True, the length is
        `len(allowable_set) + 1`.

    """
    return one_hot_encode(atom.GetTotalValence(), allowable_set,
                          include_unknown_set)



import torch
import dgl
import pytest
import numpy as np
from openpom.feat.graph_featurizer import GraphFeaturizer
from openpom.models.mpnn_pom import MPNNPOM


def test_mpnnpom_regression():
    """
    Test MPNNPOM class for regression
    """
    torch.manual_seed(0)
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    torch.set_default_device(device)
    input_smile = ["CC", "C"]
    feat = GraphFeaturizer()
    graphs = feat.featurize(input_smile)
    dgl_graphs = [graph.to_dgl_graph() for graph in graphs]
    g = dgl.batch(dgl_graphs).to(device)

    number_of_molecules = 2
    n_tasks = 3
    model = MPNNPOM(n_tasks=n_tasks,
                    mode='regression',
                    number_atom_features=134,
                    number_bond_features=6)

    # assert layers
    model.to(device)
    output = model(g)
    assert output.shape == torch.Size([number_of_molecules, n_tasks])

    required_output = np.asarray([[0.0687, 0.7174, 0.0861],
                                  [-0.1467, 0.0957, 0.1467]])
    assert np.allclose(output.detach().cpu().numpy(),
                       required_output,
                       atol=0.001)


# Set up testing parameters.
Test1_params = {
    'mpnn_residual': True,
    'message_aggregator_type': 'sum',
    'readout_type': 'set2set'
}

Test2_params = {
    'mpnn_residual': False,
    'message_aggregator_type': 'mean',
    'readout_type': 'global_sum_pooling'
}

Test3_params = {
    'mpnn_residual': True,
    'message_aggregator_type': 'max',
    'readout_type': 'set2set'
}


@pytest.mark.parametrize('test_parameters',
                         [Test1_params, Test2_params, Test3_params])
def test_mpnnpom_multiple_configs(test_parameters):
    """
    Test MPNNPOM class for multiple configs
    """
    torch.manual_seed(0)
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    torch.set_default_device(device)
    input_smile = ["CC", "C"]
    feat = GraphFeaturizer()
    graphs = feat.featurize(input_smile)
    dgl_graphs = [graph.to_dgl_graph() for graph in graphs]
    g = dgl.batch(dgl_graphs).to(device)

    number_of_molecules = 2
    n_tasks = 3

    mpnn_residual, message_aggregator_type, readout_type = \
        test_parameters.values()
    model = MPNNPOM(n_tasks=n_tasks,
                    mode='regression',
                    number_atom_features=134,
                    number_bond_features=6,
                    mpnn_residual=mpnn_residual,
                    message_aggregator_type=message_aggregator_type,
                    readout_type=readout_type)

    # assert layers
    model.to(device)
    output = model(g)
    assert output.shape == torch.Size([number_of_molecules, n_tasks])


def test_mpnnpom_classification_single_task():
    """
    Test MPNNPOM class for single task classification
    """
    torch.manual_seed(0)
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    torch.set_default_device(device)
    input_smile = ["CC", "C"]
    feat = GraphFeaturizer()
    graphs = feat.featurize(input_smile)
    dgl_graphs = [graph.to_dgl_graph() for graph in graphs]
    g = dgl.batch(dgl_graphs).to(device)

    number_of_molecules = 2
    n_tasks = 1
    n_classes = 1
    embedding_size = 2
    model = MPNNPOM(n_tasks=n_tasks,
                    mode='classification',
                    number_atom_features=134,
                    number_bond_features=6,
                    n_classes=n_classes,
                    ffn_embeddings=embedding_size)

    # assert layers
    model.to(device)
    output = model(g)
    assert len(output) == 3
    assert output[0].shape == torch.Size([number_of_molecules])
    assert output[1].shape == torch.Size([number_of_molecules, n_classes])
    assert output[2].shape == torch.Size([number_of_molecules, embedding_size])

    required_output0 = np.asarray([0.2934, 0.4143])
    required_output1 = np.asarray([[-0.8789], [-0.3463]])
    required_output2 = np.asarray([[0.3893, 0.1915], [0.0096, -0.0100]])
    assert np.allclose(output[0].detach().cpu().numpy(),
                       required_output0,
                       atol=0.001)
    assert np.allclose(output[1].detach().cpu().numpy(),
                       required_output1,
                       atol=0.001)
    assert np.allclose(output[2].detach().cpu().numpy(),
                       required_output2,
                       atol=0.001)


def test_mpnnpom_classification_multi_task():
    """
    Test MPNNPOM class for multi task classification
    """
    torch.manual_seed(0)
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    torch.set_default_device(device)
    input_smile = ["CC", "C"]
    feat = GraphFeaturizer()
    graphs = feat.featurize(input_smile)
    dgl_graphs = [graph.to_dgl_graph() for graph in graphs]
    g = dgl.batch(dgl_graphs).to(device)

    number_of_molecules = 2
    n_tasks = 3
    n_classes = 1
    embedding_size = 2
    model = MPNNPOM(n_tasks=n_tasks,
                    mode='classification',
                    number_atom_features=134,
                    number_bond_features=6,
                    n_classes=n_classes,
                    ffn_embeddings=embedding_size)

    # assert layers
    model.to(device)
    output = model(g)
    assert len(output) == 3
    assert output[0].shape == torch.Size([number_of_molecules, n_tasks])
    assert output[1].shape == torch.Size(
        [number_of_molecules, n_tasks, n_classes])
    assert output[2].shape == torch.Size([number_of_molecules, embedding_size])

    required_output0 = np.asarray([[0.2934, 0.5467, 0.3940],
                                   [0.4143, 0.6249, 0.3807]])
    required_output1 = np.asarray([[[-0.8789], [0.1873], [-0.4306]],
                                   [[-0.3463], [0.5105], [-0.4868]]])
    required_output2 = np.asarray([[0.3893, 0.1915], [0.0096, -0.0100]])
    assert np.allclose(output[0].detach().cpu().numpy(),
                       required_output0,
                       atol=0.001)
    assert np.allclose(output[1].detach().cpu().numpy(),
                       required_output1,
                       atol=0.001)
    assert np.allclose(output[2].detach().cpu().numpy(),
                       required_output2,
                       atol=0.001)



import torch
from typing import Optional, Callable, List
from deepchem.models.losses import Loss


class CustomMultiLabelLoss(Loss):
    """
    Custom Multi-Label Loss function for multi-label classification.

    The objective function is a summed cross-entropy loss over all tasks,
    with each tasks's contribution to the loss being weighted by a factor
    of log(1+ class_imbalance_ratio), such that rarer tasks were given
    a higher weighting.

    This loss function is based on:
    `A Principal Odor Map Unifies Diverse Tasks in Human Olfactory Perception
    preprint <https://www.biorxiv.org/content/10.1101/2022.09.01.504602v4>`_.

    The labels should have shape (batch_size) or (batch_size, tasks), and be
    integer class labels.  The outputs have shape (batch_size, classes) or
    (batch_size, tasks, classes) and be logits that are converted to
    probabilities using a softmax function.
    """

    def __init__(self,
                 class_imbalance_ratio: Optional[List] = None,
                 loss_aggr_type: str = 'sum',
                 device: Optional[str] = None):
        """
        Parameters
        ---------
        class_imbalance_ratio: Optional[List]
            list of class imbalance ratios
        loss_aggr_type: str
            loss aggregation type; 'sum' or 'mean'
        device: Optional[str]
            The device on which to run computations. If None, a device is
            chosen automatically.
        """
        super(CustomMultiLabelLoss, self).__init__()
        if class_imbalance_ratio is None:
            print(Warning("No class imbalance ratio provided!"))
            self.class_imbalance_ratio: Optional[torch.Tensor] = None
        else:
            self.class_imbalance_ratio = torch.Tensor(class_imbalance_ratio)

        if loss_aggr_type not in ['sum', 'mean']:
            raise ValueError(f"Invalid loss aggregate type: {loss_aggr_type}")
        self.loss_aggr_type: str = loss_aggr_type

        if device is not None:
            if self.class_imbalance_ratio is not None:
                self.class_imbalance_ratio = self.class_imbalance_ratio.to(
                    device)

    def _create_pytorch_loss(
            self) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
        """
        Returns loss function for pytorch backend
        """
        ce_loss_fn: torch.nn.CrossEntropyLoss = torch.nn.CrossEntropyLoss(
            reduction='none')

        def loss(output: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
            """
            The objective function is a summed cross-entropy loss over all
            tasks, with each tasks's contribution to the loss being weighted
            by a factor of log(1+ class_imbalance_ratio), such that rarer
            tasks were given a higher weighting.

            Parameters
            ---------
            output: torch.Tensor
                Output logits from model's forward pass per batch
            labels: torch.Tensor
                Target labels per batch

            Returns
            -------
            loss: torch.Tensor
                total or mean loss depending on loss aggregation type
            """
            # Convert (batch_size, tasks, classes)
            # to (batch_size, classes, tasks)
            # CrossEntropyLoss only supports (batch_size, classes, tasks)
            # This is for API consistency
            if len(output.shape) == 3:
                output = output.permute(0, 2, 1)

            if len(labels.shape) == len(output.shape):
                labels = labels.squeeze(-1)

            # handle multilabel
            # output shape => (batch_size, classes=1, tasks)
            # binary_output shape => (batch_size, classes=2, tasks)
            # where now we have (1 - probabilities) for ce loss calculation
            probabilities: torch.Tensor = output[:, 0, :]
            complement_probabilities: torch.Tensor = 1 - probabilities
            binary_output: torch.Tensor = torch.stack(
                [complement_probabilities, probabilities], dim=1)

            ce_loss: torch.Tensor = ce_loss_fn(binary_output, labels.long())

            if self.class_imbalance_ratio is None:
                if self.loss_aggr_type == 'sum':
                    loss: torch.Tensor = ce_loss.sum(dim=1)
                else:
                    loss = ce_loss.mean(dim=1)
            else:
                balancing_factors: torch.Tensor = torch.log(
                    1 + self.class_imbalance_ratio)

                # loss being weighted by a factor of
                # log(1+ class_imbalance_ratio)
                balanced_losses: torch.Tensor = torch.mul(
                    ce_loss, balancing_factors)

                if self.loss_aggr_type == 'sum':
                    # sum balanced loss across all tasks;
                    # shape => (batch_size)
                    loss = balanced_losses.sum(dim=1)
                else:
                    # mean balanced loss across all tasks;
                    # shape => (batch_size)
                    loss = balanced_losses.mean(dim=1)

            # duplicate loss across all tasks in a batch;
            # shape => (batch_size, n_tasks)
            # This is for API consistency
            return loss.unsqueeze(-1).repeat(1, output.shape[-1])

        return loss



import os
import numpy as np
from openpom.hyper.configs.base_config import Config


def test_config():
    """
    Test for Config base class
    """
    np.random.seed(0)
    Config.PARAMS_DICT = {
        "param1": ["p1v1", "p1v2"],
        "param2": ["p2v1", "p2v2", "p2v3"]
    }
    n_trials = 2
    param_dict, path = Config.generate_hyperparams_random(
        n_trials=n_trials, dir="openpom/hyper/test/assets")
    assert isinstance(param_dict, dict)
    assert isinstance(path, str)
    assert len(param_dict.keys()) == n_trials
    for _, params in param_dict.items():
        assert "param1" in params
        assert "param2" in params

    # remove saved file
    os.remove(path)



# yapf: disable
import numpy as np
from typing import Dict
from openpom.hyper.configs.base_config import Config


class MPNNPOMConfig(Config):
    """
    Hyperparameter search space for MPNNPOMModel

    Note:
        MPNNPOMConfig.generate_hyperparams_random() method
        might take a lot of memory during computation of
        combinations, so tweak the parameter space accordingly.
    """
    PARAMS_DICT: Dict = {
        'batch_size': [128, 256],
        'node_out_feats': list(range(50, 300, 50)),
        'edge_hidden_feats': list(range(50, 200, 25)),
        'edge_out_feats': list(range(50, 300, 50)),
        'num_step_message_passing': list(range(1, 6, 1)),
        'mpnn_residual': [True, False],
        'message_aggregator_type': ['sum', 'mean'],
        'readout_type': ['set2set', 'global_sum_pooling'],
        'num_step_set2set': [2, 3, 4, 5],
        'num_layer_set2set': [1, 2, 3],
        'ffn_hidden_list': [
                            [392],
                            [512],
                            [392, 392],
                            [512, 512],
                            [392, 392, 392],
                            [512, 512, 512],
                            ],
        'ffn_embeddings': [256],
        'ffn_dropout_p': np.linspace(0.05, 0.5, num=5).tolist(),
        'ffn_dropout_at_input_no_act': [True, False],
        'weight_decay': [0.001, 0.0001, 1e-05, 1e-06],
        'learning_rate': [0.0005, 0.001, 0.005, 0.01],
        'self_loop': [True, False],
        }



import numpy as np
from rdkit import Chem
import pytest
from openpom.feat.graph_featurizer \
    import atom_features, bond_features, GraphConvConstants, GraphFeaturizer


# Tests for helper functions
@pytest.fixture
def example_smiles_n_features():
    """
    Sample data for testing

    Returns
    -------
    dictionary
    format {'smiles':required feature vector : List}
    """
    feature_vector_C = [[
        0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0,
        0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0
    ]]
    feature_vector_NN = [[
        0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0
    ],
                         [
                             0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0,
                             0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1,
                             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                             0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0
                         ]]
    return {'C': feature_vector_C, 'N#N': feature_vector_NN}


@pytest.fixture
def example_smiles_n_b_features():
    """
    Sample data for testing

    Returns
    -------
    dictionary
    format {'smiles':required feature vector}
    """
    feature_vector_C1OC1 = [[0, 1, 0, 0, 0, 1], [0, 1, 0, 0, 0, 1],
                            [0, 1, 0, 0, 0, 1]]
    feature_vector_NN = [[0, 0, 0, 1, 0, 0]]
    return {'C1OC1': feature_vector_C1OC1, 'N#N': feature_vector_NN}


def test_atom_features_none():
    """
    Test for atom_features() with 'None' input for Atom value
    """
    f_atom = atom_features(None)
    req_f = list(np.zeros((GraphConvConstants.ATOM_FDIM, ), dtype=int))
    assert len(f_atom) == len(req_f)
    assert f_atom == req_f


def test_atom_features(example_smiles_n_features):
    """
    Test for atom_features() function
    """
    for smiles in example_smiles_n_features.keys():
        m = Chem.MolFromSmiles(smiles)
        f = []
        for atom in m.GetAtoms():
            features = atom_features(atom)
            f.append(features)
        k = np.array(f)
        req_f = np.array(example_smiles_n_features[smiles])
        assert k.shape == req_f.shape
        assert f == example_smiles_n_features[smiles]


def test_bond_features_none():
    """
    Test for bond_features() with 'None' input for bond
    """
    f_bond = bond_features(None)
    req_f = list(np.zeros((GraphConvConstants.BOND_FDIM, ), dtype=int))
    req_f[0] = 1
    assert len(f_bond) == len(req_f)
    assert f_bond == req_f


def test_bond_features(example_smiles_n_b_features):
    """
    Test for bond_features() function
    """
    for smiles in example_smiles_n_b_features.keys():
        b_f = []
        m = Chem.MolFromSmiles(smiles)
        for b in m.GetBonds():
            b_f.append(bond_features(b))
        print(b_f)
        k = np.array(b_f)
        req_f = np.array(example_smiles_n_b_features[smiles])
        assert k.shape == req_f.shape
        assert b_f == example_smiles_n_b_features[smiles]


# Tests for graph featurizer

required_edge_index = {
    "C1=CC=NC=C1":
    np.asarray([[0, 1, 1, 5, 5, 2, 2, 4, 4, 3, 3, 0],
                [1, 0, 5, 1, 2, 5, 4, 2, 3, 4, 0, 3]]),
    "CC(=O)C":
    np.asarray([[0, 3, 3, 2, 3, 1], [3, 0, 2, 3, 1, 3]]),
    "C":
    np.empty((2, 0), dtype=int)
}


def test_graph_featurizer_single_atom():
    """
    Test for featurization of "C" using `GraphFeaturizer` class.
    """
    smiles = "C"
    featurizer = GraphFeaturizer()  # is_adding_hs=False
    graph_feat = featurizer.featurize(smiles)
    assert graph_feat[0].num_nodes == 1
    assert graph_feat[0].num_node_features == GraphConvConstants.ATOM_FDIM
    assert graph_feat[0].node_features.shape == (1,
                                                 GraphConvConstants.ATOM_FDIM)
    assert graph_feat[0].num_edges == 0
    assert graph_feat[0].num_edge_features == GraphConvConstants.BOND_FDIM
    assert graph_feat[0].edge_features.shape == (0,
                                                 GraphConvConstants.BOND_FDIM)
    assert graph_feat[0].edge_index.shape == required_edge_index['C'].shape
    assert (graph_feat[0].edge_index == required_edge_index['C']).all()


def test_graph_featurizer_general_case():
    """
    Test for featurization of "CC(=O)C" using `GraphFeaturizer` class.
    """
    smiles = "CC(=O)C"
    featurizer = GraphFeaturizer()  # is_adding_hs=False
    graph_feat = featurizer.featurize(smiles)
    assert graph_feat[0].num_nodes == 4
    assert graph_feat[0].num_node_features == GraphConvConstants.ATOM_FDIM
    assert graph_feat[0].node_features.shape == (4,
                                                 GraphConvConstants.ATOM_FDIM)
    assert graph_feat[0].num_edges == 6
    assert graph_feat[0].num_edge_features == GraphConvConstants.BOND_FDIM
    assert graph_feat[0].edge_features.shape == (6,
                                                 GraphConvConstants.BOND_FDIM)
    assert graph_feat[0].edge_index.shape == required_edge_index[
        'CC(=O)C'].shape
    assert (graph_feat[0].edge_index == required_edge_index['CC(=O)C']).all()


def test_graph_featurizer_ring():
    """
    Test for featurization of "C1=CC=NC=C1" using `GraphFeaturizer` class.
    """
    smiles = "C1=CC=NC=C1"
    featurizer = GraphFeaturizer()  # is_adding_hs=False
    graph_feat = featurizer.featurize(smiles)
    assert graph_feat[0].num_nodes == 6
    assert graph_feat[0].num_node_features == GraphConvConstants.ATOM_FDIM
    assert graph_feat[0].node_features.shape == (6,
                                                 GraphConvConstants.ATOM_FDIM)
    assert graph_feat[0].num_edges == 12
    assert graph_feat[0].num_edge_features == GraphConvConstants.BOND_FDIM
    assert graph_feat[0].edge_features.shape == (12,
                                                 GraphConvConstants.BOND_FDIM)
    assert graph_feat[0].edge_index.shape == required_edge_index[
        'C1=CC=NC=C1'].shape
    assert (
        graph_feat[0].edge_index == required_edge_index['C1=CC=NC=C1']).all()



import torch
import torch.nn as nn
from typing import List, Optional, Callable, Any


class CustomPositionwiseFeedForward(nn.Module):
    """
    Customised PositionwiseFeedForward layer from deepchem
    for:
        - hidden layers of variable sizes
        - batch normalization before every activation function
        - additional output of embedding layer (penultimate layer)
          for POM embeddings.
    """

    def __init__(
        self,
        d_input: int = 1024,
        d_hidden_list: List = [1024],
        d_output: int = 1024,
        activation: str = 'leakyrelu',
        dropout_p: float = 0.0,
        dropout_at_input_no_act: bool = False,
        batch_norm: bool = True,
    ):
        """Initialize a PositionwiseFeedForward layer.

        Parameters
        ----------
        d_input: int
            Size of input layer.
        d_hidden_list: List
            List of hidden sizes.
        d_output: int (same as d_input if d_output = 0)
            Size of output layer.
        activation: str
            Activation function to be used. Can choose between 'relu' for ReLU,
            'leakyrelu' for LeakyReLU, 'prelu' for PReLU,
            'tanh' for TanH, 'selu' for SELU, 'elu' for ELU
            and 'linear' for linear activation.
        dropout_p: float
            Dropout probability.
        dropout_at_input_no_act: bool
            If true, dropout is applied on the input tensor.
            For single layer, it is not passed to an activation function.
        batch_norm: bool
            If true, applies batch normalization
            'before' every activation function
        """
        super(CustomPositionwiseFeedForward, self).__init__()

        self.dropout_at_input_no_act: bool = dropout_at_input_no_act
        self.batch_norm: bool = batch_norm

        self.activation: Callable[[Any], Any]
        if activation == 'relu':
            self.activation = nn.ReLU()

        elif activation == 'leakyrelu':
            self.activation = nn.LeakyReLU(0.1)

        elif activation == 'prelu':
            self.activation = nn.PReLU()

        elif activation == 'tanh':
            self.activation = nn.Tanh()

        elif activation == 'selu':
            self.activation = nn.SELU()

        elif activation == 'elu':
            self.activation = nn.ELU()

        elif activation == 'linear':
            self.activation = lambda x: x

        d_output = d_output if d_output != 0 else d_input

        # Set n_layers
        self.n_layers: int = len(d_hidden_list) + 1

        # Set linear layers
        if self.n_layers == 1:
            linears: List = [nn.Linear(d_input, d_output)]

        else:
            linears = [nn.Linear(d_input, d_hidden_list[0])]
            for idx in range(1, len(d_hidden_list)):
                linears.append(
                    nn.Linear(d_hidden_list[idx - 1], d_hidden_list[idx]))
            linears.append(nn.Linear(d_hidden_list[-1], d_output))

        self.linears: nn.ModuleList = nn.ModuleList(linears)
        dropout_layer: nn.Dropout = nn.Dropout(dropout_p)
        self.dropout_p: nn.ModuleList = nn.ModuleList(
            [dropout_layer for _ in range(self.n_layers)])

        if batch_norm:
            batchnorms: List = [
                nn.BatchNorm1d(d_hidden_list[idx])
                for idx in range(len(d_hidden_list))
            ]
            self.batchnorms: nn.ModuleList = nn.ModuleList(batchnorms)

    def forward(self, x: torch.Tensor) -> List[Optional[torch.Tensor]]:
        """
        Output Computation for the Customised
        PositionwiseFeedForward layer

        Parameters
        ----------
        x: torch.Tensor
            Input tensor

        Returns
        -------
        List[Optional[torch.Tensor]]
            List containing embeddings and output
        """

        if self.n_layers == 1:
            if self.dropout_at_input_no_act:
                return [None, self.linears[0](self.dropout_p[0](x))]
            else:
                return [
                    None,
                    self.dropout_p[0](self.activation(self.linears[0](x)))
                ]

        else:
            if self.dropout_at_input_no_act:
                x = self.dropout_p[-1](x)

            if self.batch_norm:
                for i in range(self.n_layers - 2):
                    x = self.dropout_p[i](self.activation(self.batchnorms[i](
                        self.linears[i](x))))

                embeddings: torch.Tensor = self.linears[self.n_layers - 2](x)
                x = self.dropout_p[self.n_layers - 2](self.activation(
                    self.batchnorms[self.n_layers - 2](embeddings)))
            else:
                for i in range(self.n_layers - 2):
                    x = self.dropout_p[i](self.activation(self.linears[i](x)))

                embeddings = self.linears[self.n_layers - 2](x)
                x = self.dropout_p[self.n_layers - 2](
                    self.activation(embeddings))

            output: torch.Tensor = self.linears[-1](x)
            return [embeddings, output]



from setuptools import setup, find_packages

# Read the requirements from the requirements.txt file
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='openpom',
    version='0.1.1',
    description='Open-source Principal Odor Map models for Olfaction',
    author='Aryan Amit Barsainyan',
    author_email='aryan.barsainyan@gmail.com',
    packages=find_packages(),
    install_requires=requirements,
)



from deepchem.models.optimizers import Optimizer
from deepchem.models.optimizers import Adam
from deepchem.models.optimizers import AdaGrad
from deepchem.models.optimizers import AdamW
from deepchem.models.optimizers import SparseAdam
from deepchem.models.optimizers import RMSProp
from deepchem.models.optimizers import GradientDescent
from deepchem.models.optimizers import KFAC


def get_optimizer(optimizer_name: str = 'adam') -> Optimizer:
    """
    Get deepchem optimizer object

    Parameters
    ---------
    optimizer_name: str
      optimizer name
      choices: [adam, adagrad, adamw, sparseadam, rmsprop, sgd, kfac]
      default: 'adam'

    Returns
    -------
    Optimizer
      Deepchem optimizer object
    """
    if optimizer_name == 'adam':
        return Adam()
    elif optimizer_name == 'adagrad':
        return AdaGrad()
    elif optimizer_name == 'adamw':
        return AdamW()
    elif optimizer_name == 'sparseadam':
        return SparseAdam()
    elif optimizer_name == 'rmsprop':
        return RMSProp()
    elif optimizer_name == 'sgd':
        return GradientDescent()
    elif optimizer_name == 'kfac':
        return KFAC()
    else:
        print("INVALID OPTIMISER NAME!, using ADAM optimizer by default")
        return Adam()



from deepchem.data.data_loader import CSVLoader
from openpom.feat.graph_featurizer import GraphFeaturizer
from openpom.utils.data_utils import get_class_imbalance_ratio


def test_class_imbalance_ratio():
    """
    Test get_class_imbalance_ratio utility on a test dataset
    """
    featurizer = GraphFeaturizer()
    smiles_field = 'smiles'
    loader = CSVLoader(tasks=['fruity', 'green', 'herbal', 'sweet', 'woody'],
                       feature_field=smiles_field,
                       featurizer=featurizer)
    input_file = \
        'openpom/utils/test/assets/test_dataset_sample_7.csv'
    dataset = loader.create_dataset(inputs=[input_file])
    class_imbalance_ratio = get_class_imbalance_ratio(dataset=dataset)
    assert class_imbalance_ratio == [1.0, 0.5, 0.5, 0.25, 0.5]
    assert class_imbalance_ratio[0] == 1.0  # max count of 'fruity'
